#define _REENTRANT  /* required for Solaris */

# if defined(__alpha) && !defined(_WIN32)
#   pragma pointer_size (save)
#   pragma pointer_size (long)
# endif

#include <sys/time.h>
#include <sys/types.h>

#ifdef __sgi
#include <sys/prctl.h>
#endif

#if !defined(AIX) && !defined(__sgi)
#include <pthread.h>
#endif /* !AIX && !__sgi */

int terminate_factorial_flag = 0; /* needed because Linux threads are separate
				     processes */

#include <string.h>

# if defined(__alpha) && !defined(_WIN32)
#   pragma pointer_size (restore)
# endif

#if !defined(__hpux) && !defined(aix)
#define PATHVAR "LD_LIBRARY_PATH"
#elif defined(aix)
#define PATHVAR "LIBPATH"
#else /* __hpux */
#define PATHVAR "SHLIB_PATH"
#endif

struct shlib_library_item
{
    char *name;
    int  system;
};

#include "shlibs.h"

static int lisp_initialized = 0, lnk_argc = 3, 
    start_semaphore[2], end_semaphore[2];
# if defined(__alpha) && !defined(_WIN32)
#   pragma pointer_size (save)
#   pragma pointer_size (long)
# endif
static char *lnk_argv[] = { "some program", "-I", "", 0 }, path_buf[1024],
    **lnk_envp = 0, dir_buf[1024];
# if defined(__alpha) && !defined(_WIN32)
#   pragma pointer_size (restore)
# endif

#if !defined(__sgi) && !defined(AIX)
static pthread_t tid;
static pthread_attr_t pattr;
#endif

void
#ifndef __sgi 
*
#endif
load_lisp( void *arg )
{
    char *slash;
    tcm( 0 
#ifdef __hpux
	 , 0
#endif
	); /* Lisp won't read from stdin and won't write to stdout/err */
    /* determine system directory */
    strcpy( dir_buf, path_buf );
    slash = strrchr( dir_buf, '/' );
    if (slash) *slash = '\0';
    lisp_init( lnk_argc, lnk_argv, lnk_envp,  /* argc, argv, envp */
	       0,  /* exit routine */
	       dir_buf, /* system directory */
#ifndef __hpux
	       "libacl503.so", /* shared library name */
#else
	       "libacl503.sl", /* shared library name */
#endif
	       "fact.dxl",  /* image name */
	       1, /* quiet flag */
	       0, /* unused on Unix */
	       linked_shared_libraries );
#ifndef __sgi
    return 0;
#endif
}

void
wait_on_semaphore( int *semaphore )
{
    fd_set fdset;
    char chr;

    FD_ZERO( &fdset );
    FD_SET( semaphore[0], &fdset );

    while (1)
    {
	if (select( semaphore[0] + 1,
#ifdef __hpux
		    (int *) /* pthread redefines select() */
#endif
		    &fdset, 0, 0, 0 ) > 0 &&
	    FD_ISSET( semaphore[0], &fdset ))
	{
	    /* clear pipe */
	    read( semaphore[0], &chr, 1 );
	    return;
	}
    }
}

void
release_semaphore( int *semaphore )
{
    char chr = 'A';

    write( semaphore[1], &chr, 1 );
    return;
}

# if defined(__alpha) && !defined(_WIN32)
#   pragma pointer_size (save)
#   pragma pointer_size (long)
# endif

int
initialize_factorial( char **envp )
# if defined(__alpha) && !defined(_WIN32)
#   pragma pointer_size (restore)
# endif
{
    lnk_envp = envp;
    if (lisp_initialized) return -1;
    if (pipe( start_semaphore ) != 0) return -1;
    if (pipe( end_semaphore ) != 0) return -1;
    /* if successful, path_buf will contain the full fact.dxl path */
    if (find_file_using_pathstring( "fact.dxl", path_buf, PATHVAR ) < 0)
    {
	return -1;
    }
    lnk_argv[2] = path_buf;
#if !defined(AIX)
#ifndef __sgi
#if !defined(__hpux)
    if (pthread_attr_init( &pattr ) != 0) return -1;
#ifndef linux
    if (pthread_attr_setstacksize( &pattr, 1024*1024 ) != 0) return -1;
#endif
#endif /* !__hpux */
    if (pthread_create( &tid, 
#ifndef __hpux
			&pattr, 
#else
			pthread_attr_default,
#endif
			load_lisp, (void *) 0) != 0) return -1;
#else /* __sgi */
    if (sproc( load_lisp, PR_SALL, 0 ) < 0) return -1;
#endif
    wait_on_semaphore( end_semaphore );
#endif /* !AIX */
    lisp_initialized = 1;
    return 0;
}

int
release_end_semaphore() /* wrapper for lisp to call when it's initialized */
{
    release_semaphore( end_semaphore );
    return 0;
}

#ifdef AIX
wait_on_end_semaphore()
{
    wait_on_semaphore( end_semaphore );
}

void *
lisp_init_func()
{
    return (void *) &load_lisp;
}
#endif /* AIX */

int (*lisp_factorial)(int, char *, int) = 0;

void
set_factorial_callback( int (*fcb) (int, char *, int) )
{
    lisp_factorial = fcb;
}

static unsigned int factorial_arg;
static int factorial_result_string_size, factorial_ret_val;
static char *factorial_result_string;

int
factorial( unsigned int x, char *result, int string_size )
{
    factorial_arg = x;
    factorial_result_string = result;
    factorial_result_string_size = string_size;
    release_semaphore( start_semaphore );
    wait_on_semaphore( end_semaphore );
    return factorial_ret_val;
}

int wait_for_work()
{
    while (1) /* loop forever */
    {
	wait_on_semaphore( start_semaphore );
	if (terminate_factorial_flag) return 0;
	factorial_ret_val = (*lisp_factorial) ( factorial_arg,
						factorial_result_string,
						factorial_result_string_size );
	release_semaphore( end_semaphore );
    }
}

/* Lisp calls this to fill result string */
int
copy_factorial_result( char *str1, char *str2, int n )
{
  strncpy( str1, str2, n - 1 );
  *(str2 + n) = '\0';
  return 0;
}

int
terminate_factorial()
{
    terminate_factorial_flag = 1;
    release_semaphore( start_semaphore );
    lisp_initialized = 0;
    return 0;
}
